""" Helpers for syncing packages into the cache in AWS Lambda """
from __future__ import print_function

import os
import sys
from distutils.spawn import find_executable  # pylint: disable=E0611,F0401

import argparse
import boto3
import json
import logging
import shutil
import six
import subprocess
import tempfile
import zipfile
from pyramid.paster import get_appsettings
from six.moves.urllib.request import urlretrieve  # pylint: disable=E0611,F0401

from pypicloud import __version__, _lambda_handler
from pypicloud.storage.s3 import S3Storage


HANDLER_FILENAME = 'lambda_script.py'

VENV_VERSION = '15.1.0'
VENV_URL = ("https://pypi.python.org/packages/source/v/"
            "virtualenv/virtualenv-%s.tar.gz" % VENV_VERSION)


def _create_role(role_name, description, policy):
    """ Idempotently create the IAM Role """
    iam = boto3.client('iam')
    iam_resource = boto3.resource('iam')
    role = iam_resource.Role(role_name)
    try:
        role.load()
    except iam.exceptions.NoSuchEntityException:
        print("Creating IAM role %s" % role_name)
        iam.create_role(
            Path='/pypicloud/',
            RoleName=role_name,
            AssumeRolePolicyDocument="""{
            "Version": "2012-10-17",
            "Statement": [{
            "Effect": "Allow",
            "Principal": {"Service": ["lambda.amazonaws.com"]},
            "Action": ["sts:AssumeRole"]
            }]}
            """,
            Description=description,
        )
        role.load()

    policy_name = "inline_policy"
    exists = False
    for role_policy in role.policies.all():
        if role_policy.name == policy_name:
            exists = True
            break
    if not exists:
        print("Attaching inline role policy")
        iam.put_role_policy(
            RoleName=role_name,
            PolicyName=policy_name,
            PolicyDocument=policy,
        )
    return role.arn


def _create_dynamodb_role(settings, bucket):
    """ Create the AWS Role needed for Lambda with a DynamoDB cache """
    # Jump through hoops to get the table names
    from pypicloud.cache.dynamo import DynamoPackage, PackageSummary
    namespace = settings.get('db.namespace', ())
    policy = """{
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": "logs:*",
                "Resource": "arn:aws:logs:*"
            },
            {
                "Effect": "Allow",
                "Action": [
                    "s3:GetObject"
                ],
                "Resource": "arn:aws:s3:::%s/*"
            },
            {
                "Effect": "Allow",
                "Action": [
                    "dynamodb:ListTables"
                ],
                "Resource": "*"
            },
            {
                "Effect": "Allow",
                "Action": [
                    "dynamodb:BatchGetItem",
                    "dynamodb:BatchWriteItem",
                    "dynamodb:DeleteItem",
                    "dynamodb:DescribeLimits",
                    "dynamodb:DescribeTable",
                    "dynamodb:GetItem",
                    "dynamodb:GetRecords",
                    "dynamodb:PutItem",
                    "dynamodb:Query",
                    "dynamodb:Scan",
                    "dynamodb:UpdateItem"
                ],
                "Resource": [%s]
            }]}
    """
    dynamo = boto3.client('dynamodb', region_name=settings['db.region_name'])
    table_arns = []
    for model in (DynamoPackage, PackageSummary):
        tablename = model.meta_.ddb_tablename(namespace)
        try:
            result = dynamo.describe_table(TableName=tablename)
        except dynamo.exceptions.ResourceNotFoundException:
            print("Could not find DynamoDB table %r. "
                  "Try running the server first to create the tables" %
                  tablename)
            sys.exit(1)
        table_arns.append(result['Table']['TableArn'])
    index_arns = [t + '/index/*' for t in table_arns]
    table_arns.extend(index_arns)
    table_arns = ['"' + t + '"' for t in table_arns]
    full_policy = policy % (bucket.name, (','.join(table_arns)))

    return _create_role(
        "pypicloud_lambda_dynamo",
        "Lambda role for syncing S3 package changes to DynamoDB",
        full_policy,
    )


def _create_default_role(settings, bucket):
    """ Create the AWS Role needed for Lambda with most caches """
    policy = """{
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": "logs:*",
                "Resource": "arn:aws:logs:*"
            },
            {
                "Effect": "Allow",
                "Action": [
                    "s3:GetObject"
                ],
                "Resource": "arn:aws:s3:::%s/*"
            }]}
    """ % bucket.name
    db = settings['pypi.db']
    return _create_role(
        "pypicloud_lambda_" + db,
        "Lambda role for syncing S3 package changes to " + db,
        policy,
    )


def make_virtualenv(env):
    """ Create a virtualenv """
    if find_executable('virtualenv') is not None:
        cmd = ['virtualenv'] + [env]
        subprocess.check_call(cmd)
    else:
        # Otherwise, download virtualenv from pypi
        path = urlretrieve(VENV_URL)[0]
        try:
            subprocess.check_call(['tar', 'xzf', path])
            subprocess.check_call(
                [sys.executable, "virtualenv-%s/virtualenv.py" % VENV_VERSION,
                 env])
        finally:
            os.unlink(path)
            shutil.rmtree("virtualenv-%s" % VENV_VERSION)


def build_lambda_bundle(argv=None):
    """ Build the zip bundle that will be deployed to AWS Lambda """
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description=build_lambda_bundle.__doc__)
    parser.add_argument('-o', help="Name of output file (default %(default)s)",
                        default='output.zip')
    parser.add_argument('config', help="Name of config file")

    args = parser.parse_args(argv)
    settings = get_appsettings(args.config)
    bundle = _build_lambda_bundle(settings)
    with open(args.o, 'wb') as ofile:
        ofile.write(bundle)
    print("Wrote bundle file %s" % args.o)


def _build_lambda_bundle(settings):
    """ Build the lambda bundle """
    venv_dir = tempfile.mkdtemp()
    try:
        print("Creating virtualenv %s" % venv_dir)
        make_virtualenv(venv_dir)
        print("Installing pypicloud into virtualenv")
        pip = os.path.join(venv_dir, 'bin', 'pip')
        subprocess.check_call([pip, 'install', "pypicloud[dynamo,redis]==%s" % __version__])

        bundle = "%s/output.zip" % venv_dir
        zipf = zipfile.ZipFile(bundle, 'w', zipfile.ZIP_DEFLATED)

        site_packages = os.path.join(venv_dir, 'lib', 'site-packages')
        if not os.path.exists(site_packages):
            subdir = [f for f in os.listdir(os.path.join(venv_dir, 'lib'))
                      if f.startswith('python')][0]
            site_packages = os.path.join(venv_dir, 'lib', subdir, 'site-packages')

        # Hack because importing some modules blows up on Lambda
        for mod in ['zope', 'repoze']:
            open(os.path.join(site_packages, mod, '__init__.py'), 'a').close()
        for root, dirs, files in os.walk(site_packages):
            for filename in files:
                fullpath = os.path.join(root, filename)
                zipf.write(fullpath, os.path.relpath(fullpath, site_packages))

        handler = os.path.join(os.path.dirname(_lambda_handler.__file__),
                               '_lambda_handler.py')
        zipf.write(handler, HANDLER_FILENAME)

        zipf.close()
        with open(bundle, 'rb') as ifile:
            return ifile.read()
    finally:
        shutil.rmtree(venv_dir)


def create_sync_scripts(argv=None):
    """
    Set bucket notifications and create AWS Lambda functions that will sync
    changes in the S3 bucket to the cache

    """
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser(description=create_sync_scripts.__doc__)
    parser.add_argument('-n', help="Name of the AWS Lambda function to "
                        " create (default %(default)s)",
                        default='pypicloud_s3_sync')
    parser.add_argument('-f', help="Overwrite existing AWS Lambda functions",
                        action='store_true')
    parser.add_argument('-a', help="The ARN of the IAM role to run as. If not "
                        "provided, pypicloud will attempt to create one.")
    parser.add_argument('-r', help="The AWS region to create the Lambda in "
                        "(will attempt to auto-detect from config file)")
    parser.add_argument('config', help="Name of config file")

    args = parser.parse_args(argv)
    logging.basicConfig()
    settings = get_appsettings(args.config)

    if args.r:
        region = args.r
    else:
        if settings.get('pypi.storage') == 's3':
            region = settings.get('storage.region_name')
    if not region:
        print("No region detected in config file. Please use -r <region>")
        sys.exit(1)

    kwargs = S3Storage.configure(settings)
    bucket = kwargs['bucket']

    # Create the Role
    if args.a:
        role_arn = args.a
    else:
        db = settings['pypi.db']
        if db == 'dynamo':
            role_arn = _create_dynamodb_role(settings, bucket)
        else:
            role_arn = _create_default_role(settings, bucket)

    lam = boto3.client('lambda', region_name=region)
    func_arn = None
    try:
        func = lam.get_function(FunctionName=args.n)
    except lam.exceptions.ResourceNotFoundException:
        pass
    else:
        if args.f:
            print("Deleting pre-existing function %r" % args.n)
            lam.delete_function(FunctionName=args.n)
        else:
            func_arn = func['Configuration']['FunctionArn']
            print("Lambda function %r already exists. Use -f to overwrite" %
                  args.n)

    # Create the lambda function
    if func_arn is None:
        bundle = _build_lambda_bundle(settings)
        handler_module = os.path.splitext(HANDLER_FILENAME)[0]
        # Pull out only the cache db settings
        small_settings = {
            'pypi.db': settings['pypi.db'],
        }
        for key, val in six.iteritems(settings):
            if key.startswith('db.'):
                small_settings[key] = val

        print("Creating Lambda function %s" % args.n)
        func = lam.create_function(
            FunctionName=args.n,
            Runtime='python2.7',
            Handler=handler_module + '.handle_s3_event',
            Code={
                'ZipFile': bundle,
            },
            Environment={
                'Variables': {
                    'PYPICLOUD_SETTINGS': json.dumps(small_settings),
                },
            },
            Description='Process S3 Object notifications & update pypicloud cache',
            Timeout=30,
            Publish=True,
            Role=role_arn,
        )
        func_arn = func['FunctionArn']

    print("Adding invoke permission for S3")
    account_id = boto3.client('sts').get_caller_identity().get('Account')
    try:
        lam.add_permission(
            Action='lambda:InvokeFunction',
            FunctionName=args.n,
            Principal='s3.amazonaws.com',
            SourceAccount=account_id,
            SourceArn="arn:aws:s3:::" + bucket.name,
            StatementId='s3_invoke',
        )
    except lam.exceptions.ResourceConflictException:
        print("Permission already present")

    print("Adding lambda configuration to S3 bucket")
    notification = bucket.Notification()
    notification.put(NotificationConfiguration={
        'LambdaFunctionConfigurations': [
            {
                'LambdaFunctionArn': func_arn,
                'Events': ['s3:ObjectCreated:*'],
            },
            {
                'LambdaFunctionArn': func_arn,
                'Events': ['s3:ObjectRemoved:*'],
            }
        ]
    })



""" Store package data in a SQL database """
import json
import logging
import transaction
import zope.sqlalchemy
from datetime import datetime
from pyramid.settings import asbool
from sqlalchemy import (engine_from_config, distinct, and_, or_, Column,
                        DateTime, String)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.mutable import Mutable
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import func
from sqlalchemy.types import TypeDecorator, TEXT

from .base import ICache
from pypicloud.models import Package


LOG = logging.getLogger(__name__)

Base = declarative_base()  # pylint: disable=C0103


class JSONEncodedDict(TypeDecorator):  # pylint: disable=W0223

    "Represents an immutable structure as a json-encoded string."

    impl = TEXT

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class MutableDict(Mutable, dict):

    """ SQLAlchemy dict field that tracks changes """

    @classmethod
    def coerce(cls, key, value):
        "Convert plain dictionaries to MutableDict."

        if not isinstance(value, MutableDict):
            if isinstance(value, dict):
                return MutableDict(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value

    def __setitem__(self, key, value):
        "Detect dictionary set events and emit change events."

        dict.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key):
        "Detect dictionary del events and emit change events."

        dict.__delitem__(self, key)
        self.changed()

MutableDict.associate_with(JSONEncodedDict)


class SQLPackage(Package, Base):

    """ Python package stored in SQLAlchemy """
    __tablename__ = 'packages'
    filename = Column(String(255, convert_unicode=True), primary_key=True)
    name = Column(String(255, convert_unicode=True), index=True, nullable=False)
    version = Column(String(1000, convert_unicode=True), nullable=False)
    last_modified = Column(DateTime(), index=True, nullable=False)
    summary = Column(String(255, convert_unicode=True), index=True, nullable=True)
    data = Column(JSONEncodedDict(), nullable=False)


def create_schema(engine):
    """
    Create the database schema if needed

    Parameters
    ----------
    engine : :class:`sqlalchemy.Engine`

    Notes
    -----
    The method should only be called after importing all modules containing
    models which extend the ``Base`` object.

    """
    Base.metadata.create_all(bind=engine)


def drop_schema(engine):
    """
    Drop the database schema

    Parameters
    ----------
    engine : :class:`sqlalchemy.Engine`

    Notes
    -----
    The method should only be called after importing all modules containing
    models which extend the ``Base`` object.

    """
    Base.metadata.drop_all(bind=engine)


class SQLCache(ICache):

    """ Caching database that uses SQLAlchemy """
    package_class = SQLPackage

    def __init__(self, request=None, dbmaker=None, graceful_reload=False, **kwargs):
        super(SQLCache, self).__init__(request, **kwargs)
        self.dbmaker = dbmaker
        self.db = self.dbmaker()
        self.graceful_reload = graceful_reload

        if request is not None:
            zope.sqlalchemy.register(self.db, transaction_manager=request.tm)

    def reload_if_needed(self):
        super(SQLCache, self).reload_if_needed()
        if self.request is None:
            self.db.commit()
            self.db.close()

    @classmethod
    def configure(cls, settings):
        kwargs = super(SQLCache, cls).configure(settings)
        graceful_reload = asbool(settings.pop('db.graceful_reload', False))
        engine = engine_from_config(settings, prefix='db.')
        # Create SQL schema if not exists
        create_schema(engine)
        kwargs['dbmaker'] = sessionmaker(bind=engine)
        kwargs['graceful_reload'] = graceful_reload
        return kwargs

    def fetch(self, filename):
        return self.db.query(SQLPackage).filter_by(filename=filename).first()

    def all(self, name):
        pkgs = self.db.query(SQLPackage).filter_by(name=name).all()
        pkgs.sort(reverse=True)
        return pkgs

    def distinct(self):
        names = self.db.query(distinct(SQLPackage.name))\
            .order_by(SQLPackage.name).all()
        return [n[0] for n in names]

    def search(self, criteria, query_type):
        """
        Perform a search.

        Queries are performed as follows:

            For the AND query_type, queries within a column will utilize the
            AND operator, but will not conflict with queries in another column.

                (column1 LIKE '%a%' AND column1 LIKE '%b%')
                OR
                (column2 LIKE '%c%' AND column2 LIKE '%d%')

            For the OR query_type, all queries will utilize the OR operator:

                (column1 LIKE '%a%' OR column1 LIKE '%b%')
                OR
                (column2 LIKE '%c%' OR column2 LIKE '%d%')

        """
        conditions = []
        for key, queries in criteria.items():
            # Make sure search key exists in the package class.
            # It should be either "name" or "summary".
            field = getattr(self.package_class, key, None)
            if not field:
                continue

            column_conditions = []

            for query in queries:
                # Generate condition and add to list
                condition = field.like('%' + query + '%')
                column_conditions.append(condition)

            # Check if conditions for this column were generated
            if column_conditions:
                if query_type == 'and':
                    operator = and_
                else:
                    operator = or_
                conditions.append(operator(*column_conditions))

        # Piece together the queries. Refer to the method docstring for
        # examples as to how this works.
        results = self.db.query(SQLPackage).filter(or_(*conditions))

        # Extract only the most recent version for each package
        latest_map = {}
        for package in results.all():
            if package.name not in latest_map or \
               package > latest_map[package.name]:
                latest_map[package.name] = package

        return latest_map.values()

    def summary(self):
        subquery = self.db.query(
            SQLPackage.name,
            func.max(SQLPackage.last_modified).label('last_modified'),
        ).group_by(SQLPackage.name).subquery()
        rows = self.db.query(
            SQLPackage.name,
            SQLPackage.last_modified,
            SQLPackage.summary,
        ).filter(
            (SQLPackage.name == subquery.c.name) &
            (SQLPackage.last_modified == subquery.c.last_modified)
        )

        # Dedupe because two packages may share the same last_modified
        seen_packages = set()
        packages = []
        for row in rows:
            if row[0] in seen_packages:
                continue
            seen_packages.add(row[0])
            packages.append({
                'name': row[0],
                'last_modified': row[1],
                'summary': row[2],
            })
        return packages

    def clear(self, package):
        self.db.delete(package)

    def clear_all(self):
        # Release any transactions before we go reloading schema
        if self.request is None:
            self.db.rollback()
        else:
            transaction.abort()
        engine = self.dbmaker.kw['bind']
        drop_schema(engine)
        create_schema(engine)

    def save(self, package):
        self.db.merge(package)

    def reload_from_storage(self):
        if not self.graceful_reload:
            return super(SQLCache, self).reload_from_storage()

        LOG.info("Rebuilding cache from storage")
        # Log start time
        start = datetime.utcnow()
        # Fetch packages from storage s1
        s1 = set(self.storage.list(self.package_class))
        # Fetch cache packages c1
        c1 = set(self.db.query(SQLPackage).all())
        # Add missing packages to cache (s1 - c1)
        missing = s1 - c1
        if missing:
            LOG.info("Adding %d missing packages to cache", len(missing))
            for pkg in missing:
                self.db.merge(pkg)
        # Delete extra packages from cache (c1 - s1) when last_modified < start
        # The time filter helps us avoid deleting packages that were
        # concurrently uploaded.
        extra1 = [p for p in (c1 - s1) if p.last_modified <= start]
        if extra1:
            LOG.info("Removing %d extra packages from cache", len(extra1))
            for pkg in extra1:
                self.db.query(SQLPackage) \
                    .filter(SQLPackage.filename == pkg.filename) \
                    .delete(synchronize_session=False)

        # If any packages were concurrently deleted during the cache rebuild,
        # we can detect them by polling storage again and looking for any
        # packages that were present in s1 and are missing from s2
        s2 = set(self.storage.list(self.package_class))
        # Delete extra packages from cache (s1 - s2)
        extra2 = s1 - s2
        if extra2:
            LOG.info("Removing %d packages from cache that were concurrently "
                     "deleted during rebuild", len(extra2))
            for pkg in extra2:
                self.db.query(SQLPackage) \
                    .filter(SQLPackage.filename == pkg.filename) \
                    .delete(synchronize_session=False)



""" AWS Lambda handler that process S3 object notifications """
from __future__ import print_function

import os
import posixpath

import boto3
import json
from datetime import datetime


def handle_s3_event(event, context):
    """ Handle S3 object notification """
    from pypicloud.cache import get_cache_impl
    from pypicloud.storage.s3 import package_from_object
    from pypicloud.util import parse_filename

    settings = json.loads(os.environ['PYPICLOUD_SETTINGS'])
    # Set 'file' storage as a hack. We're going to load the cache, which will
    # load a storage. We won't actually be using the storage for anything, but
    # the settings have to be present.
    settings.setdefault('pypi.storage', 'file')
    settings.setdefault('storage.dir', '/tmp')
    cache_impl = get_cache_impl(settings)
    kwargs = cache_impl.configure(settings)
    cache = cache_impl(**kwargs)

    s3 = boto3.resource('s3')
    for record in event['Records']:
        bucket = record['s3']['bucket']['name']
        key = record['s3']['object']['key']
        event_name = record['eventName']
        if event_name.startswith('ObjectCreated'):
            print("S3 object %r created" % key)
            obj = s3.Object(bucket, key)
            package = package_from_object(obj, cache.package_class)
            existing_pkg = cache.fetch(package.filename)
            if existing_pkg is None:
                print("Saving package %s" % package)
                cache.save(package)
            else:
                print("Package already cached")
        else:
            print("S3 object %r deleted" % key)
            filename = posixpath.basename(key)
            try:
                name, version = parse_filename(filename)
            except ValueError:
                name = version = 'dummy'
            package = cache.package_class(name, version, filename, datetime.utcnow(), '')
            print("Deleting package %s" % package)
            cache.clear(package)



""" S3-backed pypi server """
import os
import sys

import calendar
import datetime
import io
import logging
import traceback
from pyramid.config import Configurator
from pyramid.renderers import JSON, render
from pyramid.settings import asbool
from pyramid_beaker import session_factory_from_settings
from six.moves.urllib.parse import urlencode  # pylint: disable=F0401,E0611

from .route import Root
from .util import BetterScrapingLocator


__version__ = '1.0.2'
LOG = logging.getLogger(__name__)


def to_json(value):
    """ A json filter for jinja2 """
    return render('json', value)


def _app_url(request, *paths, **params):
    """ Get the base url for the root of the app plus an optional path """
    path = '/'.join(paths)
    if not path.startswith('/'):
        path = '/' + path
    if params:
        path += '?' + urlencode(params)
    return request.application_url + path


def _locator(request):
    """ Get the scraping locator to find packages from the fallback site """
    return BetterScrapingLocator(request.registry.fallback_url)


def includeme(config):
    """ Set up and configure the pypicloud app """
    config.set_root_factory(Root)
    settings = config.get_settings()
    config.add_route('health', '/health')
    config.include('pyramid_tm')
    # Beaker should be set by default to invalidate corrupt sessions, otherwise
    # a bad cookie will break the website for you and the only fix is to
    # manually delete the cookie.
    settings.setdefault('session.invalidate_corrupt', 'true')
    config.include('pyramid_beaker')
    config.include('pyramid_duh')
    config.include('pyramid_duh.auth')
    config.include('pyramid_rpc.xmlrpc')
    config.include('pypicloud.auth')
    config.include('pypicloud.access')
    config.include('pypicloud.cache')

    # If we're reloading templates, we should also pretty-print json
    reload_templates = asbool(settings.get('pyramid.reload_templates'))
    indent = 4 if reload_templates else None
    json_renderer = JSON(indent=indent)
    json_renderer.add_adapter(datetime.datetime, lambda obj, r:
                              calendar.timegm(obj.utctimetuple()))
    config.add_renderer('json', json_renderer)
    # Jinja2 configuration
    settings['jinja2.filters'] = {
        'static_url': 'pyramid_jinja2.filters:static_url_filter',
        'tojson': to_json,
    }
    settings['jinja2.directories'] = ['pypicloud:templates']
    config.include('pyramid_jinja2')

    # BEAKER CONFIGURATION
    config.registry.secure_cookie = asbool(settings.get('session.secure',
                                                        False))
    settings.setdefault('session.type', 'cookie')
    settings.setdefault('session.httponly', 'true')
    config.set_session_factory(session_factory_from_settings(settings))

    # PYPICLOUD SETTINGS
    default_url = 'https://pypi.python.org/simple'
    config.registry.fallback_url = settings.get('pypi.fallback_url',
                                                default_url)

    fallback_mode = settings.get('pypi.fallback', 'redirect')
    always_show_upstream = settings.get('pypi.always_show_upstream')

    # Using fallback=mirror is the same as fallback=cache and
    # always_show_upstream=true
    if always_show_upstream is None:
        always_show_upstream = fallback_mode == 'mirror'
    else:
        always_show_upstream = asbool(always_show_upstream)

    modes = ('redirect', 'cache', 'none')
    if fallback_mode not in modes:
        raise ValueError("Invalid value for 'pypi.fallback'. "
                         "Must be one of %s" % ', '.join(modes))
    config.registry.fallback = fallback_mode
    config.registry.always_show_upstream = always_show_upstream

    # Special request methods
    config.add_request_method(_app_url, name='app_url')
    config.add_request_method(_locator, name='locator', reify=True)
    config.add_request_method(lambda x: __version__, name='pypicloud_version',
                              reify=True)
    config.add_request_method(lambda x: settings.get('pypi.download_url'),
                              name='custom_download_url', reify=True)

    cache_max_age = int(settings.get('pyramid.cache_max_age', 3600))
    config.add_static_view(name='static/%s' % __version__,
                           path='pypicloud:static',
                           cache_max_age=cache_max_age)

    config.add_xmlrpc_endpoint('pypi', '/pypi', request_method='POST',
                               header='Content-Type:text/xml')
    config.add_xmlrpc_endpoint('pypi_slash', '/pypi/', request_method='POST',
                               header='Content-Type:text/xml')


def traceback_formatter(excpt, value, tback):
    """ Catches all exceptions and re-formats the traceback raised.
    """

    sys.stdout.write("".join(traceback.format_exception(excpt, value, tback)))


def hook_exceptions():
    """ Hooks into the sys module to set our formatter.
    """

    if hasattr(sys.stdout, "fileno"):  # when testing, sys.stdout is StringIO
        try:
            # reopen stdout in non buffered mode
            sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)
        except (io.UnsupportedOperation, ValueError):
            pass
        else:
            # set the hook
            sys.excepthook = traceback_formatter


def main(config, **settings):
    """ This function returns a Pyramid WSGI application.
    """
    hook_exceptions()
    config = Configurator(settings=settings)
    config.include('pypicloud')
    config.scan('pypicloud.views')
    return config.make_wsgi_app()



"""LDAP authentication plugin for pypicloud."""
import logging
from collections import namedtuple
from functools import wraps
from pyramid.settings import aslist, asbool

from .base import IAccessBackend
from pypicloud.util import TimedCache


try:
    import ldap
except ImportError:  # pragma: no cover
    raise ImportError(
        "You must 'pip install pypicloud[ldap]' before using ldap as the "
        "authentication backend."
    )


LOG = logging.getLogger(__name__)


def reconnect(func):
    """
    If the LDAP connection dies underneath us, recreate it
    """
    @wraps(func)
    def _reconnect(self, *args, **kwargs):
        """
        Inner wrap function to reconnect on failure
        """
        try:
            return func(self, *args, **kwargs)
        except ldap.LDAPError:
            self.connect()
            return func(*args, **kwargs)

    return _reconnect


User = namedtuple('User', ['username', 'dn', 'is_admin'])


class LDAP(object):
    """ Handles interactions with the remote LDAP server """

    def __init__(self, admin_field, admin_value, base_dn, cache_time,
                 service_dn, service_password, service_username, url,
                 user_search_filter, user_dn_format, ignore_cert):
        self._url = url
        self._service_dn = service_dn
        self._service_password = service_password
        self._base_dn = base_dn
        self._user_search_filter = user_search_filter
        self._user_dn_format = user_dn_format
        if user_dn_format is not None:
            if base_dn is not None or user_search_filter is not None:
                raise ValueError("Cannot use user_dn_format with base_dn "
                                 "and user_search_filter")
        else:
            if base_dn is None or user_search_filter is None:
                raise ValueError("Must provide user_dn_format or both base_dn "
                                 "and user_search_filter")
        self._admin_field = admin_field
        self._admin_value = admin_value
        self._server = None
        if cache_time is not None:
            cache_time = int(cache_time)
        self._cache = TimedCache(cache_time, self._fetch_user)
        if service_username is not None:
            self._cache.set_expire(
                service_username,
                User(service_username, service_dn, True),
                None
            )
        self._ignore_cert = ignore_cert

    def connect(self):
        """ Initializes the python-ldap module and does the initial bind """
        if self._ignore_cert:
            ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)
        LOG.debug("LDAP connecting to %s", self._url)
        self._server = ldap.initialize(self._url)
        self._bind_to_service()

    def _bind_to_service(self):
        """ Bind to the service account or anonymous """
        if self._service_dn:
            # bind with the service_dn
            self._server.simple_bind_s(self._service_dn, self._service_password)
        else:
            # force a connection without binding
            self._server.whoami_s()

    @reconnect
    def _fetch_user(self, username):
        """ Fetch a user entry from the LDAP server """
        LOG.debug("LDAP fetching user %s", username)
        search_attrs = []
        if self._admin_field is not None:
            search_attrs.append(self._admin_field)
        if self._user_dn_format is not None:
            dn = self._user_dn_format.format(username=username)
            LOG.debug("LDAP searching user %r with dn %r", username, dn)
            results = self._server.search_s(dn, ldap.SCOPE_BASE,
                                            attrlist=search_attrs)
        else:
            search_filter = self._user_search_filter.format(username=username)
            LOG.debug("LDAP searching user %r with filter %r", username,
                      search_filter)
            results = self._server.search_s(
                self._base_dn,
                ldap.SCOPE_SUBTREE,
                search_filter,
                search_attrs,
            )
        if not results:
            LOG.debug("LDAP user %r not found", username)
            return None
        if len(results) > 1:
            raise ValueError("More than one user found for %r: %r" %
                             (username, [r[0] for r in results]))
        dn, attributes = results[0]

        is_admin = False
        if self._admin_field is not None:
            if self._admin_field in attributes:
                is_admin = any((val in attributes[self._admin_field]
                                for val in self._admin_value))

        return User(username, dn, is_admin)

    def get_user(self, username):
        """ Get the User object or None """
        return self._cache.get(username)

    @reconnect
    def verify_user(self, username, password):
        """
        Attempts to bind as the user, then rebinds as service user again
        """
        LOG.debug("LDAP verifying user %s", username)
        # Empty password may successfully complete an anonymous bind.
        # Explicitly disallow empty passwords.
        if password == "":
            return False
        user = self._cache.get(username)
        if user is None:
            return False

        try:
            LOG.debug("LDAP binding user %r", user.dn)
            self._server.simple_bind_s(user.dn, password)
        except ldap.INVALID_CREDENTIALS:
            return False
        else:
            return True
        finally:
            self._bind_to_service()


class LDAPAccessBackend(IAccessBackend):
    """
    This backend allows you to authenticate against a remote LDAP server.
    """
    def __init__(self, request=None, conn=None, **kwargs):
        super(LDAPAccessBackend, self).__init__(request, **kwargs)
        self.conn = conn

    @classmethod
    def configure(cls, settings):
        kwargs = super(LDAPAccessBackend, cls).configure(settings)
        conn = LDAP(
            admin_field=settings.get('auth.ldap.admin_field'),
            admin_value=aslist(settings.get('auth.ldap.admin_value', [])),
            base_dn=settings.get('auth.ldap.base_dn'),
            cache_time=settings.get('auth.ldap.cache_time'),
            service_dn=settings.get('auth.ldap.service_dn'),
            service_password=settings.get('auth.ldap.service_password', ''),
            service_username=settings.get('auth.ldap.service_username'),
            url=settings['auth.ldap.url'],
            user_dn_format=settings.get('auth.ldap.user_dn_format'),
            user_search_filter=settings.get('auth.ldap.user_search_filter'),
            ignore_cert=asbool(settings.get('auth.ldap.ignore_cert'))
        )
        conn.connect()
        kwargs['conn'] = conn
        return kwargs

    def _get_password_hash(self, *_):  # pragma: no cover
        raise RuntimeError("LDAP should never call _get_password_hash")

    def verify_user(self, username, password):
        return self.conn.verify_user(username, password)

    def groups(self, username=None):
        # We're not supporting groups for LDAP
        return []  # pragma: no cover

    def group_members(self, group):
        # We're not supporting groups for LDAP
        return []  # pragma: no cover

    def is_admin(self, username):
        if not username:
            return False
        user = self.conn.get_user(username)
        return user is not None and user.is_admin

    def group_permissions(self, package):
        return {}

    def user_permissions(self, package):
        return {}

    def user_package_permissions(self, username):
        return []

    def group_package_permissions(self, group):
        return []

    def user_data(self, username=None):
        if username is None:
            return []
        else:
            return {
                "username": username,
                "admin": self.is_admin(username),
                "groups": self.groups(username),
            }



""" Store packages in S3 """
import posixpath
from binascii import hexlify

import boto3
import logging
import rsa
from botocore.config import Config
from botocore.signers import CloudFrontSigner
from botocore.exceptions import ClientError
from contextlib import contextmanager
from datetime import datetime, timedelta
from hashlib import md5
from pyramid.httpexceptions import HTTPFound
from pyramid.settings import asbool, falsey
from pyramid_duh.settings import asdict
from six.moves.urllib.parse import quote  # pylint: disable=F0401,E0611
from six.moves.urllib.request import urlopen  # pylint: disable=F0401,E0611

from .base import IStorage
from pypicloud.models import Package
from pypicloud.util import parse_filename, get_settings


LOG = logging.getLogger(__name__)


def package_from_object(obj, factory):
    """ Create a package from a S3 object """
    filename = posixpath.basename(obj.key)
    name = obj.metadata.get('name')
    version = obj.metadata.get('version')
    summary = obj.metadata.get('summary')
    # We used to not store metadata. This is for backwards
    # compatibility
    if name is None or version is None:
        try:
            name, version = parse_filename(filename)
        except ValueError:
            LOG.warning("S3 file %s has no package name", obj.key)
            return None

    return factory(name, version, filename, obj.last_modified, summary,
                   path=obj.key)


class S3Storage(IStorage):

    """ Storage backend that uses S3 """
    test = False

    def __init__(self, request=None, bucket=None, expire_after=None,
                 bucket_prefix=None, prepend_hash=None, redirect_urls=None,
                 sse=None, object_acl=None,
                 **kwargs):
        super(S3Storage, self).__init__(request, **kwargs)
        self.bucket = bucket
        self.expire_after = expire_after
        self.bucket_prefix = bucket_prefix
        self.prepend_hash = prepend_hash
        self.redirect_urls = redirect_urls
        self.sse = sse
        self.object_acl = object_acl

    @classmethod
    def configure(cls, settings):
        kwargs = super(S3Storage, cls).configure(settings)
        kwargs['expire_after'] = int(settings.get('storage.expire_after',
                                                  60 * 60 * 24))
        kwargs['bucket_prefix'] = settings.get('storage.prefix', '')
        kwargs['prepend_hash'] = asbool(settings.get('storage.prepend_hash',
                                                     True))
        kwargs['sse'] = sse = settings.get('storage.server_side_encryption')
        if sse not in [None, 'AES256', 'aws:kms']:
            LOG.warn("Unrecognized value %r for 'storage.sse'. See "
                     "https://boto3.readthedocs.io/en/latest/reference/services/s3.html#S3.Object.put "
                     "for more details", sse)
        kwargs['object_acl'] = settings.get('storage.object_acl', None)
        kwargs['redirect_urls'] = asbool(settings.get('storage.redirect_urls',
                                                      False))

        config_settings = get_settings(
            settings,
            'storage.',
            region_name=str,
            signature_version=str,
            user_agent=str,
            user_agent_extra=str,
            connect_timeout=int,
            read_timeout=int,
            parameter_validation=asbool,
            max_pool_connections=int,
            proxies=asdict,
        )
        config_settings['s3'] = get_settings(
            settings,
            'storage.',
            use_accelerate_endpoint=asbool,
            payload_signing_enabled=asbool,
            addressing_style=str,
        )
        config = Config(**config_settings)

        def verify_value(val):
            """ Verify can be a boolean (False) or a string """
            s = str(val).strip().lower()
            if s in falsey:
                return False
            else:
                return str(val)

        s3conn = boto3.resource(
            's3',
            config=config,
            **get_settings(
                settings,
                'storage.',
                region_name=str,
                api_version=str,
                use_ssl=asbool,
                verify=verify_value,
                endpoint_url=str,
                aws_access_key_id=str,
                aws_secret_access_key=str,
                aws_session_token=str,
            )
        )

        bucket_name = settings.get('storage.bucket')
        if bucket_name is None:
            raise ValueError("You must specify the 'storage.bucket'")
        bucket = s3conn.Bucket(bucket_name)
        try:
            bucket.load()
        except ClientError as e:
            if e.response['Error']['Code'] == '404':
                LOG.info("Creating S3 bucket %s", bucket_name)
                bucket.create()
                bucket.wait_until_exists()
            else:
                if e.response['Error']['Code'] == '301':
                    LOG.warn("Bucket found in different region. Check that "
                             "the S3 bucket specified in 'storage.bucket' is "
                             "in 'storage.region_name'")
                raise
        kwargs['bucket'] = bucket
        return kwargs

    def calculate_path(self, package):
        """ Calculates the path of a package """
        path = package.name + '/' + package.filename
        if self.prepend_hash:
            m = md5()
            m.update(package.filename.encode('utf-8'))
            prefix = hexlify(m.digest()).decode('utf-8')[:4]
            path = prefix + '/' + path
        return path

    def get_path(self, package):
        """ Get the fully-qualified bucket path for a package """
        if 'path' not in package.data:
            filename = self.calculate_path(package)
            package.data['path'] = self.bucket_prefix + filename
        return package.data['path']

    def list(self, factory=Package):
        keys = self.bucket.objects.filter(Prefix=self.bucket_prefix)
        for summary in keys:
            # ObjectSummary has no metadata, so we have to fetch it.
            obj = summary.Object()
            pkg = package_from_object(obj, factory)
            if pkg is not None:
                yield pkg

    def _generate_url(self, package):
        """ Generate a signed url to the S3 file """
        return self.bucket.meta.client.generate_presigned_url(
            'get_object',
            Params={
                'Bucket': self.bucket.name,
                'Key': self.get_path(package),
            },
            ExpiresIn=self.expire_after,
        )

    def get_url(self, package):
        if self.redirect_urls:
            return super(S3Storage, self).get_url(package)
        else:
            return self._generate_url(package)

    def download_response(self, package):
        return HTTPFound(location=self._generate_url(package))

    def upload(self, package, datastream):
        key = self.bucket.Object(self.get_path(package))
        kwargs = {}
        if self.sse is not None:
            kwargs['ServerSideEncryption'] = self.sse
        if self.object_acl:
            kwargs['ACL'] = self.object_acl
        metadata = {
            'name': package.name,
            'version': package.version,
        }
        if package.summary:
            metadata['summary'] = package.summary
        key.put(Metadata=metadata, Body=datastream, **kwargs)

    def delete(self, package):
        self.bucket.delete_objects(Delete={
            'Objects': [
                {
                    'Key': self.get_path(package),
                }
            ]
        })

    @contextmanager
    def open(self, package):
        url = self._generate_url(package)
        handle = urlopen(url)
        try:
            yield handle
        finally:
            handle.close()


class CloudFrontS3Storage(S3Storage):

    """ Storage backend that uses S3 and CloudFront """
    def __init__(self, request=None, domain=None, private_key=None,
                 key_id=None, **kwargs):
        super(CloudFrontS3Storage, self).__init__(request, **kwargs)
        self.domain = domain
        self.private_key = private_key
        self.key_id = key_id
        self.private_key = private_key

        self.cf_signer = None
        if key_id is not None:
            self.cf_signer = CloudFrontSigner(self.key_id, self._rsa_signer)

        self.client = boto3.client('cloudfront')

    @classmethod
    def configure(cls, settings):
        kwargs = super(CloudFrontS3Storage, cls).configure(settings)
        kwargs['domain'] = settings['storage.cloud_front_domain']
        kwargs['key_id'] = settings.get('storage.cloud_front_key_id')
        private_key = settings.get('storage.cloud_front_key_string')
        if private_key is None:
            key_file = settings.get('storage.cloud_front_key_file')
            if key_file:
                with open(key_file, 'r') as ifile:
                    private_key = ifile.read()
        kwargs['private_key'] = private_key

        return kwargs

    def _rsa_signer(self, message):
        """ Generate a RSA signature for a message """
        return rsa.sign(
            message,
            rsa.PrivateKey.load_pkcs1(self.private_key.encode('utf8')),
            'SHA-1')  # CloudFront requires SHA-1 hash

    def _generate_url(self, package):
        """ Get the fully-qualified CloudFront path for a package """
        path = self.get_path(package)
        url = self.domain + '/' + quote(path)

        # No key id, no signer, so we don't have to sign the URL
        if self.cf_signer is None:
            return url

        # To sign with a canned policy:
        expires = datetime.utcnow() + timedelta(seconds=self.expire_after)
        return self.cf_signer.generate_presigned_url(
            url, date_less_than=expires)
